@@ -1,11 +1,14 @@
 package Test::Compile;
+
 use 5.006;
 use warnings;
 use strict;
+
 use Test::Builder;
 use File::Spec;
 use UNIVERSAL::require;
-our $VERSION = '0.13_1';
+
+our $VERSION = '0.16_1';
 my $Test = Test::Builder->new;
 
 sub import {
@@ -28,10 +31,10 @@ sub pm_file_ok {
     my $file = shift;
     my $name = @_ ? shift : "Compile test for $file";
 
-    my ($ok,$diag) = _check_syntax($file,1);
+    my $ok = _run_in_subprocess(sub{_check_syntax($file,1)});
 
     $Test->ok($ok, $name);
-    $Test->diag($diag) unless $ok;
+    $Test->diag("$file does not compile") unless $ok;
     return $ok;
 }
 
@@ -53,10 +56,10 @@ sub pl_file_ok {
         }
     }
 
-    my ($ok,$diag) = _check_syntax($file,0);
+    my $ok = _run_in_subprocess(sub{_check_syntax($file,0)});
 
     $Test->ok($ok, $name);
-    $Test->diag($diag) unless $ok;
+    $Test->diag("$file does not compile") unless $ok;
     return $ok;
 }
 
@@ -72,6 +75,7 @@ sub all_pm_files_ok {
 
 sub all_pl_files_ok {
     my @files = @_ ? @_ : all_pl_files();
+    $Test->skip_all("no pl files found") unless @files;
     $Test->plan(tests => scalar @files);
     my $ok = 1;
     for (@files) {
@@ -97,7 +101,7 @@ sub all_pl_files {
 
     my @pl;
     for my $file ( _find_files(@queue) ) {
-        if (-f $file) {
+        if (defined($file) && -f $file) {
             # Only accept files with no extension or extension .pl
             push @pl, $file if $file =~ /(?:^[^.]+$|\.pl$)/;
         }
@@ -105,42 +109,49 @@ sub all_pl_files {
     return @pl;
 }
 
+sub _run_in_subprocess {
+    my ($closure) = @_;
+
+    my $pid = fork();
+    if ( ! defined($pid) ) {
+        return 0;
+    } elsif ( $pid ) {
+        wait();
+        return ($? ? 0 : 1);
+    } else {
+        close(STDERR);
+        my $rv = $closure->();
+        exit ($rv ? 0 : 1);
+    }
+}
+
 sub _check_syntax {
     my ($file,$require) = @_;
 
-    if (! -f $file) {
-        return (0,"$file does not exist");
-    }
-
-    if ( $require ) {
-        my $module = $file;
-        $module =~ s!^(blib[/\\])?lib[/\\]!!;
-        $module =~ s![/\\]!::!g;
-        $module =~ s/\.pm$//;
-
-        $module->use;
-        if ( $@ ) {
-            return (0,"couldn't use $module ($file): $@");
-        }
-    } else {
-        open my $olderr, ">&", \*STDERR or die "Can't dup STDERR: $!";
-        close(STDERR);
-        system($^X, (map { "-I$_" } split ':', $ENV{PERL5LIB}), '-c', $file);
-        my $error = $?;
-        open STDERR, ">&", $olderr or die "Can't dup olderr: $!";
-        if ($error) {
-            return (0,"Script does not compile");
+    if (-f $file) {
+        if ( $require ) {
+            my $module = $file;
+            $module =~ s!^(blib[/\\])?lib[/\\]!!;
+            $module =~ s![/\\]!::!g;
+            $module =~ s/\.pm$//;
+    
+            $module->use;
+            return ($@ ? 0 : 1);
+        } else {
+            my @perl5lib = split(':', ($ENV{PERL5LIB}||""));
+            my $taint = _is_in_taint_mode($file);
+            unshift @perl5lib, 'blib/lib';
+            system($^X, (map { "-I$_" } @perl5lib), "-c$taint", $file);
+            return ($? ? 0 : 1);
         }
     }
-
-    return (1,undef);
 }
 
 sub _find_files {
     my (@queue) = @_;
 
     for my $file (@queue) {
-        if (-d $file) {
+        if (defined($file) && -d $file) {
             local *DH;
             opendir DH, $file or next;
             my @newfiles = readdir DH;
@@ -169,6 +180,19 @@ sub _pl_starting_points {
     return 'script' if -e 'script';
     return 'bin'    if -e 'bin';
 }
+
+sub _is_in_taint_mode {
+    my $file = shift;
+    open(FILE, $file) or die "could not open $file";
+    my $shebang = <FILE>;
+    my $taint = "";
+    if ($shebang =~ /^#![\/\w]+\s+\-w?([tT])/) {
+        $taint = $1;
+    }
+    close FILE;
+    return $taint;
+}
+
 1;
 __END__
 
@@ -339,7 +363,7 @@ Evan Giles, C<< <egiles@cpan.org> >>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright 2007-2011 by the authors.
+Copyright 2007-2012 by the authors.
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself.
